from .schema import AtomType
from .visitor import Visitor


def template(schema):
    return Templater().template_node(schema)


class Templater(Visitor):
    def template_node(self, value):
        data = self.visit(value)

        if value.optional:
            if isinstance(data, dict):
                data.update(value.default)
            elif isinstance(data, (float, int, str)):
                data = value.default

        return data

    def visit_tuple(self, tup):
        return [
            self.template_node(node)
            for node in tup.elements
        ]

    def visit_record(self, rec):
        return {
            key: self.template_node(node)
            for key, node in rec.fields.items()
        }

    def visit_union(self, union):
        return self.template_node(union.options[0])

    def visit_atom(self, atom):
        if atom.type_ == AtomType.OPTION:
            return f"... {atom.options}"
        elif atom.type_ == AtomType.BOOL:
            return "... bool"
        elif atom.type_ == AtomType.FLOAT:
            return "... float"
        elif atom.type_ == AtomType.INT:
            return "... int"
        elif atom.type_ == AtomType.STR:
            return "... string"

    def visit_map(self, map):
        return {
            "XX": self.template_node(map.element)
        }

    def visit_collection(self, collection):
        return [
            self.template_node(collection.element)
        ]



from .schema import Node


class Visitor:
    def visit(self, node: Node, *args, **kwargs):
        return node.structure.accept(self, *args, **kwargs)

    def visit_atom(self, atom, *args, **kwargs):
        pass

    def visit_union(self, union, *args, **kwargs):
        for node in union.options:
            self.visit(node, *args, **kwargs)

    def visit_record(self, rec, *args, **kwargs):
        for _, node in rec.fields.items():
            self.visit(node, *args, **kwargs)

    def visit_collection(self, collection, *args, **kwargs):
        self.visit(collection.element)

    def visit_map(self, map_, *args, **kwargs):
        self.visit(map_.element)

    def visit_tuple(self, tup, *args, **kwargs):
        for node in tup.fields:
            self.visit(node, *args, **kwargs)



import pytest

from .. import AtomType, Node, validate
from ..validate import InvalidDataError


def test_atom():
    u = Node.of_atom(AtomType.INT)
    assert validate(45, u) == 45


def test_union():
    u = Node.of_union([Node.of_atom(AtomType.INT), Node.of_atom(AtomType.FLOAT)])

    assert validate(45, u) == (Node.of_atom(AtomType.INT), 45)
    assert validate(3.14, u) == (Node.of_atom(AtomType.FLOAT), 3.14)


def test_collection_valid():
    u = Node.of_collection(Node.of_atom(AtomType.INT))
    assert validate([1, 2], u) == [1, 2]


def test_collection_invalid_item():
    u = Node.of_collection(Node.of_atom(AtomType.INT))
    with pytest.raises(InvalidDataError):
        validate([1, 2, 4.6], u)


def test_collection_str():
    u = Node.of_collection(Node.of_atom(AtomType.STR))
    with pytest.raises(InvalidDataError):
        validate("abc", u)


def test_record():
    u = Node.of_record(
        {
            "a": Node.of_atom(AtomType.INT),
            "b": Node.of_atom(AtomType.FLOAT, optional=True, default=45.0),
        }
    )

    assert validate({"a": 78}, u) == {"a": 78, "b": 45.0}

    with pytest.raises(InvalidDataError):
        validate({"b": 78}, u)


def test_collection_of_union():
    schema = Node.of_collection(
        Node.of_union(
            [
                Node.of_atom(AtomType.INT),
                Node.of_atom(AtomType.FLOAT),
            ]
        )
    )

    assert validate([1, 2, 4.6], schema) == [
        (Node.of_atom(AtomType.INT), 1),
        (Node.of_atom(AtomType.INT), 2),
        (Node.of_atom(AtomType.FLOAT), 4.6),
    ]
    with pytest.raises(InvalidDataError):
        validate([1, 2, "foo", 4.6], schema)



from dataclasses import dataclass
from typing import List

from .result import Result

from .schema import AtomType, Atom, Node, Record, Collection
from .visitor import Visitor
from .misc import closest


def validate(data, schema, fail=True):
    """Validate a piece of data against a Schema and return an interepreted result."""
    assert isinstance(schema, Node), f"schema should be a Node instance, not {schema}."

    r = Validator.validate(schema, data)

    if r.valid:
        return r.value
    elif fail:
        raise InvalidDataError(r.error)

    return None


class Validator(Visitor):
    @classmethod
    def validate(cls, schema, data):
        """validate data against schema.

        :param schema: instance of Node
        :param data: the data to validate
        """
        return cls().visit(schema, data)

    def visit_atom(self, atom, data) -> Result:
        if atom.type_ == AtomType.INT:
            if isinstance(data, int):
                return Result.ok(data)
            elif isinstance(data, float):
                if int(data) == data:
                    return Result.ok(int(data))

        elif atom.type_ == AtomType.FLOAT:
            if isinstance(data, (float, int)):
                return Result.ok(data)

        elif atom.type_ == AtomType.STR:
            if isinstance(data, str):
                return Result.ok(data)

        elif atom.type_ == AtomType.OPTION:
            assert atom.options is not None, "atom.options must be set"
            if isinstance(data, str) and data in atom.options:
                return Result.ok(data)

        elif atom.type_ == AtomType.BOOL:
            if isinstance(data, bool):
                return Result.ok(data)

        else:
            raise NotImplementedError(f"{atom.type_} is not supported yet.")

        # invalid atom
        return Result.fail(
            Mismatch(
                [],
                data,
                f"{repr(data)} is not a valid {atom.type_.name(atom.options)}.",
            )
        )

    def visit_union(self, union, data) -> Result:
        errors = []
        for op in union.options:
            m = self.visit(op, data)
            if m.valid:
                # success at first match
                return Result.ok((op, m.value))
            else:
                errors.append(m.error)

        # no valid option
        return Result.fail(
            Mismatch(
                [],
                data,
                f"{repr(data)} does not validate any of the valid alternatives:\n"
                + "\n".join("- " + error_message(err, no_loc=True) for err in errors),
            )
        )

    def visit_record(self, rec, data) -> Result:
        try:
            pairs = data.items()
        except AttributeError:
            # early return because we cannot go further anyway
            return Result.fail(
                Mismatch([], data, f"{repr(data)} does not implement the items method.")
            )

        got = set()
        res = {}
        errors = []

        for key, value in pairs:
            if key not in rec.fields:
                other = closest(rec.fields.keys(), key)
                errors.append(
                    Mismatch([], data, f"{key} is not a valid member of this record. Did you mean {other}")
                )
                continue

            m = self.visit(rec.fields[key], value)
            if m.valid:
                got.add(key)
                res[key] = m.value
            else:
                errors.append(m.error.from_(key))

        for key in set(rec.fields.keys()).difference(got):
            if not rec.fields[key].optional:
                # collect missing members
                errors.append(Mismatch([], data, f"{key} is missing."))
            else:
                default = rec.fields[key].default
                r = self.visit(rec.fields[key], default)
                if not r.valid:
                    return Result.fail(
                        Mismatch([key], default, "Default value is invalid data.")
                    )
                res[key] = r.value

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)

    def visit_tuple(self, tup, data) -> Result:
        try:
            elements = list(data)
        except TypeError:
            # early return because we cannot go further anyway
            return Result.fail(Mismatch([], data, f"{repr(data)} is not iterable."))

        res = []
        errors = []

        if len(elements) != len(tup.fields):
            errors.append(
                Mismatch(
                    [],
                    data,
                    f"Expected {len(tup.fields)} elements, but got {len(elements)}.",
                )
            )

        for i, (schema, value) in enumerate(zip(tup.fields, elements)):
            m = self.visit(schema, value)
            if m.valid:
                res.append(m.value)
            else:
                errors.append(m.error.from_(str(i)))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)

    def visit_collection(self, collection, data) -> Result:
        if isinstance(data, str):
            return Result.fail(Mismatch([], data, f"{repr(data)} is a string, not a collection"))

        try:
            col = list(data)
        except AttributeError:
            return Result.fail(Mismatch([], data, f"{repr(data)} is not iterable."))

        validate = []
        errors = []

        for i, e in enumerate(col):
            m = self.visit(collection.element, e)
            if m.valid:
                validate.append(m.value)
            else:
                # collect mismatches in elements
                errors.append(m.error.from_(str(i)))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(validate)

    def visit_map(self, map_, data) -> Result:
        try:
            pairs = data.items()
        except AttributeError:
            return Result.fail(
                Mismatch([], data, f"{repr(data)} does not implement the items method.")
            )

        res = {}
        errors = []

        for key, value in pairs:
            m = self.visit(map_.element, value)
            if m.valid:
                res[key] = m.value
            else:
                # collect mismatches in elements
                errors.append(m.error.from_(key))

        if errors:
            return Result.fail(Mismatches(errors))
        else:
            return Result.ok(res)


@dataclass
class Mismatch:
    where: List[str]
    what: object
    why: str

    def from_(self, where):
        return Mismatch([where, *self.where], self.what, self.why)


class Mismatches:
    def __init__(self, errors):
        self.errors = []

        for err in errors:
            if isinstance(err, Mismatches):
                self.errors.extend(err.errors)
            else:
                self.errors.append(err)

    def from_(self, where):
        return Mismatches([e.from_(where) for e in self.errors])


class InvalidDataError(ValueError):
    def __init__(self, mismatch):
        super().__init__(error_message(mismatch))


def error_message(mismatch, no_loc=False):
    if isinstance(mismatch, Mismatches):
        msg = []
        for mism in mismatch.errors:
            msg.append(error_message(mism, no_loc=no_loc))

        return "\n".join(msg)
    elif no_loc:
        return mismatch.why
    else:
        if None:
            location = ".".join(mismatch.where)
        else:
            location = "[toplevel]"

        return f"In {location}: {mismatch.why}"



from .schema import Node, Atom, AtomType, Record, Tuple
from .visitor import Visitor
from .misc import closest, lev_dist


def describe(schema, selects=()):
    res = select(schema, selects)

    if not res:
        i = res.err

        if i == 0:
            raise KeyError(f"There is no {selects[0]} field in this schema.")
        else:
            raise KeyError(
                f"There is no {selects[i]} field in " + ".".join(selects[:i]) + "."
            )

    if selects != ():
        print(".".join(selects) + ":")

    for s in Describer.describe(res.val):
        print(s)


def select(schema, selects, index=0):
    if selects in {("all",), ("",)}:
        return Result.ok(schema)

    if index < len(selects) - 1:
        if (
            isinstance(schema.structure, (Record, Tuple))
            and selects[index] in schema.structure.fields
        ):
            node = schema.structure.fields[selects[index]]
            return select(node, selects, index=index + 1)
        else:
            return Result.error(index)
    else:
        return Result.ok(schema.structure.fields[selects[index]])


class Describer(Visitor):
    @classmethod
    def describe(cls, schema):
        """Show the visit"""
        yield from cls().describe_node(schema)

    def visit_atom(self, atom, indent=0):
        indent_str = "  " * indent
        message = [f"{indent_str}"]

        if atom.type_ == AtomType.OPTION:
            if len(atom.options) == 1:
                message.append(repr(atom.options[0]))
            else:
                message.append("one of ")
                message.append(", ".join(map(repr, atom.options[:-1])))
                message.append(f", or {repr(atom.options[-1])}")
        else:
            message.append(f"value of type {atom.type_.name()}")

        yield "".join(message)

    def visit_union(self, union, indent=0):
        indent_str = "  " * indent

        if len(union.options) > 1:
            yield indent_str + "one of the following options:"
            pfx = indent_str + "- "
            ind = indent + 1
        else:
            pfx = ""
            ind = indent

        for op in union.options:
            yield from self.describe_node(op, prefix=pfx, indent=ind)

    def visit_record(self, rec, indent=0):
        indent_str = "  " * indent
        for key, node in rec.fields.items():
            yield from self.describe_node(
                node, prefix=f"{indent_str}{key}: ", indent=indent + 1,
            )

    def visit_tuple(self, tup, indent=0):
        indent_str = "  " * indent
        for value in tup.elements:
            yield from self.describe_node(
                value, prefix=f"{indent_str}- ", indent=indent + 1,
            )

    def visit_collection(self, collection, indent=0):
        indent_str = "  " * indent
        yield f"{indent_str}a list of{indent_str}"
        yield from self.visit(collection.element, indent=indent + 1)

    def visit_map(self, map, indent=0):
        yield from self.visit(
            map.element, indent=indent + 1
        )

    def describe_node(self, value, prefix="", indent=0):
        indent_str = "  " * indent
        message = [prefix]

        if value.optional:
            message.append(
                f"(optional, default is {repr(value.default)})"
            )

        if value.description:
            if value.optional:
                message.append("\n" + indent_str)

            indented_descr = []

            first, *rest = value.description.split("\n")

            indented_descr.append(first)

            for line in rest:
                indented_descr.append(indent_str + line)

            message.append("\n".join(indented_descr))

        msg = "".join(message)

        if msg:
            yield msg

        yield from self.visit(value, indent=indent + 1)


class Result:
    def __init__(self, valid, val, err):
        self.val = val
        self.err = err
        self.valid = valid

    @classmethod
    def error(cls, err):
        return cls(False, None, err)

    @classmethod
    def ok(cls, val):
        return cls(True, val, None)

    def __bool__(self):
        return self.valid


def search(schema, name):
    likely = sorted(Searcher.search(schema, name))
    return [".".join(p) for _, p in likely]


class Searcher(Visitor):
    @classmethod
    def search(cls, schema, name):
        """Show the visit"""
        yield from cls().visit(schema, name)

    def visit_atom(self, atom, name, path=()):
        return ()

    def visit_union(self, union, name, path=()):
        for op in union.options:
            yield from self.visit(op, name, path)

    def visit_record(self, rec, name, path=()):
        for key, node in rec.fields.items():
            yield (lev_dist(key, name), (*path, key))
            yield from self.visit(node, name, (*path, key))

    def visit_tuple(self, tup, name, path=()):
        for i, value in enumerate(tup.fields):
            yield from self.visit(
                value,
                name,
                (*path, i),
            )

    def visit_collection(self, collection, name, path=()):
        yield from self.visit(collection.element, name, path)

    def visit_map(self, map, name, path=()):
        yield from self.visit(map.element, name, path)



from dataclasses import dataclass
from typing import Optional, List, Dict, Any
from enum import Enum
from abc import abstractmethod


class Structure:
    """Abstract structure of a node"""

    @abstractmethod
    def accept(self, visitor, *args, **kwargs):
        pass

    @classmethod
    def structure_default(cls):
        raise ValueError(
            f"In {cls} it is an error to set the node optional without a default value."
        )


@dataclass
class Node:
    """Base class for a node of the schema"""

    structure: Structure
    description: Optional[str] = None
    optional: bool = False
    default: Any = None

    def __init__(
        self,
        structure: Structure,
        description: Optional[str] = None,
        optional: bool = False,
        default: Any = None,
    ):
        self.structure = structure
        self.description = description
        self.optional = optional
        self.default = default

        if self.optional and self.default is None:
            self.default = self.structure.structure_default()

    @classmethod
    def of_atom(cls, atom, options=None, **kwargs):
        return cls(Atom(atom, options=options), **kwargs)

    @classmethod
    def of_union(cls, options, **kwargs):
        return cls(Union(options), **kwargs)

    @classmethod
    def of_collection(cls, element, **kwargs):
        return cls(Collection(element), **kwargs)

    @classmethod
    def of_record(cls, fields, **kwargs):
        return cls(Record(fields), **kwargs)

    @classmethod
    def of_map(cls, element, **kwargs):
        return cls(Map(element), **kwargs)

    @classmethod
    def of_tuple(cls, elements, **kwargs):
        return cls(Tuple(elements), **kwargs)


class AtomType(Enum):
    INT = 0
    FLOAT = 1
    STR = 2
    BOOL = 3
    OPTION = 4

    def name(self, options=None):
        if self == self.INT:
            return "int"
        elif self == self.FLOAT:
            return "float"
        elif self == self.BOOL:
            return "bool"
        elif self == self.STR:
            return "str"
        elif self == self.OPTION:
            return "option(" + ", ".join(repr(s) for s in options) + ")"
        else:
            raise NotImplementedError("This type is not yet named...")


@dataclass
class Atom(Structure):
    """Structure of an atomic piece of data"""

    type_: AtomType
    options: Optional[List[str]] = None

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_atom(self, *args, **kwargs)


@dataclass
class Union(Structure):
    """Structure for the union of several nodes"""

    options: List[Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_union(self, *args, **kwargs)


@dataclass
class Collection(Structure):
    """An ordered collection of similar nodes"""

    element: Node

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_collection(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return []


@dataclass
class Record(Structure):
    """A key-value pair collection.

    Keys are from a defined set and values are defined per-key.
    Keys are supposed to be valid identifiers.
    """

    fields: Dict[str, Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_record(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return {}


@dataclass
class Map(Structure):
    """A key-value pair collection.

    Keys are not restricted and values are of a single type.
    """

    element: Node

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_map(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return {}


@dataclass
class Tuple(Structure):
    """A fixed set of ordered fields.

    Each field can have a different structure.
    """

    fields: List[Node]

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_tuple(self, *args, **kwargs)

    @classmethod
    def structure_default(cls):
        return []



def closest(alternatives, key):
    return min(alternatives, key=lambda s: lev_dist(key, s))


def lev_dist(a, b):
    len_a = len(a) + 1
    len_b = len(b) + 1

    if len_b > len_a:
        return lev_dist(b, a)

    prev = list(range(len_a))
    current = [0] * len_a

    for j in range(1, len_b):
        current[0] = j
        for i in range(1, len_a):
            if a[i - 1] == b[j - 1]:
                current[i] = prev[i - 1]  # mat[i, j] = mat[i - 1, j - 1]
            else:
                current[i] = min(  # mat[i, j] =
                    current[i - 1] + 1,  # deletion mat[i - 1, j] + 1
                    prev[i] + 1,  # insertion mat[i, j - 1]
                    prev[i - 1] + 1,  # substitution  mat[i - 1, j - 1]
                )

        # swap rows
        current, prev = prev, current

    return prev[-1]



class Result:
    def __init__(self, valid: bool, val):
        if valid:
            self._value = val
            self._error = None
        else:
            self._value = None
            self._error = val

        self.valid = valid

    @classmethod
    def ok(cls, value):
        return cls(True, value)

    @classmethod
    def fail(cls, error):
        return cls(False, error)

    @property
    def value(self):
        assert self.valid, "Try to unwrap invalid result."
        return self._value

    @property
    def error(self):
        assert not self.valid, "Try to unwrap the error of a valid result."
        return self._error

    def __bool__(self):
        return self.valid



from .. import Node, AtomType, validate


def test_record():

    data = {
        "title": "DFE for Diamond",
        "savename": "Diamond",
        "name": "Diamond",
        "main_dir": ".",
        "compet_dir": "compet",
        "host_dir": "host/supercell",
        "defects_dir": "defects",
        "dfe": {
            "me": 0.9174,
            "mh": 1.2899,
            "xmin": 0.0,
            "xmax": 6.03,
            "ymin": 0.0,
            "ymax": 12.0,
        },
        "conc": {
            "xmin": 800.0,
            "xmax": 1500.0,
            "ymin": 1.0,
            "ymax": 1e25,
            "dt": 30.0,
            "troom": 300.0,
        },
        "plot": {"show": False, "save_figure_fmt": "png"},
        "corrections": {
            "params": "params.csv",
            "bs": {"dEC": 0.92, "dEV": -0.99},
            "mp": {"eps": 5.66, "geom": "fcc", "mk": 1.9065},
        },
    }

    schema = Node.of_record(
        {
            "title": Node.of_atom(AtomType.STR, description="The title of the plot"),
            "savename": Node.of_atom(AtomType.STR, description="The name of the saved file"),
            "name": Node.of_atom(AtomType.STR, description="The name of the study"),
            "main_dir": Node.of_atom(
                AtomType.STR,
                description="Root dir of the study, all path are relative to this one",
            ),
            "compet_dir": Node.of_atom(
                AtomType.STR,
                description="Path to the competition phases relative to <main_dir>",
            ),
            "host_dir": Node.of_atom(
                AtomType.STR,
                description="Path to OUTCAR and DOSCAR files of the host supercell relative to <main_dir>",
            ),
            "defects_dir": Node.of_atom(
                AtomType.STR, description="Where to find the defects relative to <main_dir>"
            ),
            "q_subdir": Node.of_atom(AtomType.STR, optional=True, default="."),
            "load_store_cells": Node.of_atom(
                AtomType.BOOL, optional=True, default=False
            ),
            "load_store_mstudies": Node.of_atom(
                AtomType.BOOL, optional=True, default=False
            ),
            "export_study": Node.of_atom(AtomType.BOOL, optional=True, default=True),
            "dfe": Node.of_record(
                {
                    "me": Node.of_atom(
                        AtomType.FLOAT,
                        description="Effective mass of the electrons in the conductance band",
                    ),
                    "mh": Node.of_atom(
                        AtomType.FLOAT,
                        description="Effective mass of the holes in the valence band",
                    ),
                    "xmin": Node.of_atom(AtomType.FLOAT, description="Minimum of mu_Ef [eV]"),
                    "xmax": Node.of_atom(AtomType.FLOAT, description="Maximum of mu_Ef [eV]"),
                    "ymin": Node.of_atom(AtomType.FLOAT, description="Maximum of DFE [eV]"),
                    "ymax": Node.of_atom(AtomType.FLOAT, description="Maximum of DFE [eV]"),
                }
            ),
            "conc": Node.of_record(
                {
                    "xmin": Node.of_atom(
                        AtomType.FLOAT, description="Minimum of T_growth [K]"
                    ),
                    "xmax": Node.of_atom(
                        AtomType.FLOAT, description="Maximum of T_growth [K]"
                    ),
                    "ymin": Node.of_atom(
                        AtomType.FLOAT,
                        description="Minimum concentrations [cm-3] !!! The + is required in YAML scientific notation",
                    ),
                    "ymax": Node.of_atom(
                        AtomType.FLOAT, description="Maximum concentrations [cm-3]"
                    ),
                    "dt": Node.of_atom(AtomType.FLOAT, description="Delte T_growth"),
                    "troom": Node.of_atom(AtomType.FLOAT),
                }
            ),
            "plot": Node.of_record(
                {
                    "show": Node.of_atom(AtomType.BOOL),
                    "save_figure_fmt": Node.of_atom(AtomType.STR),
                }
            ),
            "corrections": Node.of_record(
                {
                    "params": Node.of_atom(
                        AtomType.STR,
                        description="Name of the .csv containing the different correction parameters",
                    ),
                    "bs": Node.of_record(
                        {
                            "apply": Node.of_atom(
                                AtomType.BOOL, optional=True, default=True
                            ),
                            "dEC": Node.of_atom(
                                AtomType.FLOAT,
                                description="Shift between CBM at the main functional and CBM at the precise functional",
                            ),
                            "dEV": Node.of_atom(
                                AtomType.FLOAT,
                                description="Shift between VBM at the main functional and VBM at the precise functional",
                            ),
                            # "gap": Node.of_atom(AtomType.OPTION, options=[float])
                        },
                        description="Band edge shift",
                    ),
                    "mb": Node.of_record(
                        {
                            "apply": Node.of_atom(
                                AtomType.BOOL, optional=True, default=True
                            )
                        },
                        description="Moss-Burstein corrections",
                        optional=True,
                    ),
                    "mp": Node.of_record(
                        {
                            "apply": Node.of_atom(
                                AtomType.BOOL, optional=True, default=True
                            ),
                            "eps": Node.of_atom(
                                AtomType.FLOAT,
                                description="Isotropic static dielectric tensor",
                            ),
                            "geom": Node.of_atom(
                                AtomType.OPTION,
                                options=["fcc", "bcc", "sc", "hcp", "other"],
                            ),
                            "mk": Node.of_atom(
                                AtomType.FLOAT,
                                description="In the case q=1 & e_r=1, equals to TEWEN for H in box",
                            ),
                        },
                        description="Makov-Payne corrections",
                    ),
                },
                description="Semicolon based CSV. Path relative to <main_dir>",
            ),
        }
    )

    assert validate(data, schema)
